@@ -35,6 +35,7 @@
 #include <fcitx/keys.h>
 #include <fcitx/module.h>
 #include <fcitx/context.h>
+#include <fcitx/module/punc/punc.h>
 #include <string>
 #include <libintl.h>
 
@@ -65,7 +66,9 @@ boolean LoadSunpinyinConfig(FcitxSunpinyinConfig* fs);
 static void SaveSunpinyinConfig(FcitxSunpinyinConfig* fs);
 static void ConfigSunpinyin(FcitxSunpinyin* sunpinyin);
 static void* SunpinyinGetFullPinyin(void* arg, FcitxModuleFunctionArg args);
+static void* SunpinyinAddWord(void* arg, FcitxModuleFunctionArg args);
 static INPUT_RETURN_VALUE FcitxSunpinyinDeleteCandidate (FcitxSunpinyin* sunpinyin, FcitxCandidateWord* candWord);
+static void UpdatePunc(FcitxSunpinyin* sunpinyin);
 
 
 static const char* fuzzyPairs[][2] = {
@@ -98,11 +101,15 @@ static const char *correctionPairs[][2] = {
  * @brief Reset the status.
  *
  **/
-__EXPORT_API
 void FcitxSunpinyinReset (void* arg)
 {
     FcitxSunpinyin* sunpinyin = (FcitxSunpinyin*) arg;
     sunpinyin->view->clearIC();
+    FcitxUIStatus* puncStatus = FcitxUIGetStatusByName(sunpinyin->owner, "punc");
+    if (puncStatus)
+        sunpinyin->view->setStatusAttrValue(CIMIWinHandler::STATUS_ID_FULLPUNC, puncStatus->getCurrentStatus(puncStatus->arg));
+    else
+        sunpinyin->view->setStatusAttrValue(CIMIWinHandler::STATUS_ID_FULLPUNC, false);
 }
 
 /**
@@ -113,7 +120,6 @@ void FcitxSunpinyinReset (void* arg)
  * @param count count from XKeyEvent
  * @return INPUT_RETURN_VALUE
  **/
-__EXPORT_API
 INPUT_RETURN_VALUE FcitxSunpinyinDoInput(void* arg, FcitxKeySym sym, unsigned int state)
 {
     FcitxSunpinyin* sunpinyin = (FcitxSunpinyin*) arg;
@@ -122,12 +128,13 @@ INPUT_RETURN_VALUE FcitxSunpinyinDoInput(void* arg, FcitxKeySym sym, unsigned in
     FcitxWindowHandler* windowHandler = sunpinyin->windowHandler;
     FcitxSunpinyinConfig* fs = &sunpinyin->fs;
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(sunpinyin->owner);
-    FcitxCandidateWordSetChoose(FcitxInputStateGetCandidateList(input), DIGIT_STR_CHOOSE);
+    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+    FcitxCandidateWordSetChoose(candList, DIGIT_STR_CHOOSE);
 
     int chooseKey = FcitxHotkeyCheckChooseKey(sym, FcitxKeyState_None, DIGIT_STR_CHOOSE);
     if (state == FcitxKeyState_Ctrl_Alt && chooseKey >= 0)
     {
-        FcitxCandidateWord* candidateWord = FcitxCandidateWordGetByIndex(FcitxInputStateGetCandidateList(input), chooseKey);
+        FcitxCandidateWord* candidateWord = FcitxCandidateWordGetByIndex(candList, chooseKey);
         return FcitxSunpinyinDeleteCandidate(sunpinyin, candidateWord);
     }
 
@@ -150,9 +157,10 @@ INPUT_RETURN_VALUE FcitxSunpinyinDoInput(void* arg, FcitxKeySym sym, unsigned in
         return IRV_TO_PROCESS;
 
     if (FcitxHotkeyIsHotKey(sym, state, FCITX_SPACE))
-        return FcitxCandidateWordChooseByIndex(FcitxInputStateGetCandidateList(input), 0);
+        return FcitxCandidateWordChooseByIndex(candList, 0);
 
-    if (!FcitxHotkeyIsHotKeyUAZ(sym, state)
+    if ((view->getIC()->isEmpty() || !sunpinyin->fs.bProcessPunc)
+        && !FcitxHotkeyIsHotKeyUAZ(sym, state)
         && !FcitxHotkeyIsHotKeyLAZ(sym, state)
         && !FcitxHotkeyIsHotKey(sym, state, FCITX_SEMICOLON)
         && !FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE)
@@ -193,6 +201,7 @@ boolean FcitxSunpinyinInit(void* arg)
 {
     FcitxSunpinyin* sunpinyin = (FcitxSunpinyin* )arg;
     FcitxInstanceSetContext(sunpinyin->owner, CONTEXT_IM_KEYBOARD_LAYOUT, "us");
+    UpdatePunc(sunpinyin);
     return true;
 }
 
@@ -203,27 +212,27 @@ boolean FcitxSunpinyinInit(void* arg)
  * @param searchMode
  * @return INPUT_RETURN_VALUE
  **/
-__EXPORT_API
 INPUT_RETURN_VALUE FcitxSunpinyinGetCandWords(void* arg)
 {
     FcitxSunpinyin* sunpinyin = (FcitxSunpinyin* )arg;
     FcitxInstance* instance = sunpinyin->owner;
     FcitxInputState* input = FcitxInstanceGetInputState(instance);
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(sunpinyin->owner);
-    FcitxCandidateWordSetPageSize(FcitxInputStateGetCandidateList(input), config->iMaxCandWord);
+    FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+    FcitxCandidateWordSetPageSize(candList, config->iMaxCandWord);
 
     CPreEditString ppd;
     sunpinyin->view->getPreeditString(ppd);
     TIConvSrcPtr src = (TIConvSrcPtr) (ppd.string());
-    
-    int hzlen = 0;    
+
+    int hzlen = 0;
     while (hzlen < ppd.charTypeSize())
     {
         if ((ppd.charTypeAt(hzlen) & IPreeditString::HANZI_CHAR) != IPreeditString::HANZI_CHAR)
             break;
         hzlen ++ ;
     }
-    
+
     FcitxInstanceCleanInputWindowUp(instance);
 
     memcpy(sunpinyin->front_src, src, ppd.caret() * sizeof(TWCHAR));
@@ -244,9 +253,9 @@ INPUT_RETURN_VALUE FcitxSunpinyinGetCandWords(void* arg)
     WCSTOMBS(sunpinyin->preedit, sunpinyin->front_src, MAX_USER_INPUT);
     FcitxInputStateSetCursorPos(input, strlen(sunpinyin->preedit));
     WCSTOMBS(&sunpinyin->preedit[strlen(sunpinyin->preedit)], sunpinyin->end_src, MAX_USER_INPUT);
-    
+
     FcitxInputStateSetShowCursor(input, true);
-    
+
     FcitxMessagesAddMessageAtLast(FcitxInputStateGetPreedit(input), MSG_INPUT, "%s", sunpinyin->preedit);
 
     CCandidateList pcl;
@@ -272,13 +281,10 @@ INPUT_RETURN_VALUE FcitxSunpinyinGetCandWords(void* arg)
         candWord.strWord = strdup(sunpinyin->ubuf);
         candWord.wordType = MSG_OTHER;
 
-        FcitxCandidateWordAppend(FcitxInputStateGetCandidateList(input), &candWord);
+        FcitxCandidateWordAppend(candList, &candWord);
 
         if (i == 0)
-        {
             FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", candWord.strWord);
-        }
-
     }
     return IRV_DISPLAY_CANDWORDS;
 }
@@ -289,7 +295,6 @@ INPUT_RETURN_VALUE FcitxSunpinyinGetCandWords(void* arg)
  * @param iIndex index of candidate word
  * @return the string of canidate word
  **/
-__EXPORT_API
 INPUT_RETURN_VALUE FcitxSunpinyinGetCandWord (void* arg, FcitxCandidateWord* candWord)
 {
     FcitxSunpinyin* sunpinyin = (FcitxSunpinyin* )arg;
@@ -313,7 +318,6 @@ INPUT_RETURN_VALUE FcitxSunpinyinGetCandWord (void* arg, FcitxCandidateWord* can
  * @param arg
  * @return successful or not
  **/
-__EXPORT_API
 void* FcitxSunpinyinCreate (FcitxInstance* instance)
 {
     FcitxSunpinyin* sunpinyin = (FcitxSunpinyin*) fcitx_utils_malloc0(sizeof(FcitxSunpinyin));
@@ -327,7 +331,7 @@ void* FcitxSunpinyinCreate (FcitxInstance* instance)
         free(sunpinyin);
         return NULL;
     }
-    
+
 #if FCITX_CHECK_VERSION(4,2,1)
     /* portable detect here */
     if (getenv("FCITXDIR")) {
@@ -337,7 +341,7 @@ void* FcitxSunpinyinCreate (FcitxInstance* instance)
         free(path);
     }
 #endif
-    
+
     CSunpinyinSessionFactory& fac = CSunpinyinSessionFactory::getFactory();
 
     if (fs->bUseShuangpin)
@@ -364,6 +368,8 @@ void* FcitxSunpinyinCreate (FcitxInstance* instance)
     sunpinyin->windowHandler->SetOwner(sunpinyin);
     sunpinyin->view->setStatusAttrValue(CIMIWinHandler::STATUS_ID_FULLSYMBOL, false);
     sunpinyin->view->setStatusAttrValue(CIMIWinHandler::STATUS_ID_FULLPUNC, false);
+    sunpinyin->puncOp = new CGetFullPunctOp;
+    sunpinyin->view->getIC()->setGetFullPunctOp(sunpinyin->puncOp);
     ConfigSunpinyin(sunpinyin);
 
     FcitxInstanceRegisterIM(instance,
@@ -376,14 +382,15 @@ void* FcitxSunpinyinCreate (FcitxInstance* instance)
                     FcitxSunpinyinDoInput,
                     FcitxSunpinyinGetCandWords,
                     NULL,
-                    NULL,
+                    FcitxSunpinyinSave,
                     ReloadConfigFcitxSunpinyin,
                     NULL,
-                    fs->iSunpinyinPriority,
+                    1,
                     "zh_CN"
                    );
 
     AddFunction(addon, (void*) SunpinyinGetFullPinyin);
+    AddFunction(addon, (void*) SunpinyinAddWord);
 
     return sunpinyin;
 }
@@ -393,7 +400,6 @@ void* FcitxSunpinyinCreate (FcitxInstance* instance)
  *
  * @return int
  **/
-__EXPORT_API
 void FcitxSunpinyinDestroy (void* arg)
 {
     FcitxSunpinyin* sunpinyin = (FcitxSunpinyin*) arg;
@@ -418,11 +424,19 @@ INPUT_RETURN_VALUE FcitxSunpinyinDeleteCandidate (FcitxSunpinyin* sunpinyin, Fci
         CIMIClassicView* classicView = (CIMIClassicView*) sunpinyin->view;
         unsigned int mask;
         classicView->deleteCandidate(*index, mask);
+        classicView->updateWindows(mask);
         return IRV_DISPLAY_CANDWORDS;
     }
     return IRV_TO_PROCESS;
 }
 
+void FcitxSunpinyinSave(void* arg)
+{
+    FcitxSunpinyin* sunpinyin = (FcitxSunpinyin*) arg;
+    sunpinyin->view->getIC()->memorize();
+}
+
+
 /**
  * @brief Load the config file for fcitx-sunpinyin
  *
@@ -473,6 +487,7 @@ void ConfigSunpinyin(FcitxSunpinyin* sunpinyin)
                 prof->addPageDownKey(CKeyEvent(config->hkNextPage[i].sym, 0, config->hkNextPage[i].state));
         }
         sunpinyin->view->setCancelOnBackspace(1);
+        sunpinyin->view->getIC()->setHistoryPower(fs->iMemoryStrength);
     }
 
     string_pairs fuzzy, correction;
@@ -512,9 +527,10 @@ void ConfigSunpinyin(FcitxSunpinyin* sunpinyin)
     AShuangpinSchemePolicy::instance().setShuangpinType(fs->SPScheme);
     AQuanpinSchemePolicy::instance().setFuzzySegmentation(fs->bFuzzySegmentation);
     AQuanpinSchemePolicy::instance().setInnerFuzzySegmentation(fs->bFuzzyInnerSegmentation);
+    UpdatePunc(sunpinyin);
 }
 
-__EXPORT_API void ReloadConfigFcitxSunpinyin(void* arg)
+void ReloadConfigFcitxSunpinyin(void* arg)
 {
     FcitxSunpinyin* sunpinyin = (FcitxSunpinyin*) arg;
     LoadSunpinyinConfig(&sunpinyin->fs);
@@ -554,4 +570,77 @@ void* SunpinyinGetFullPinyin(void* arg, FcitxModuleFunctionArg args)
         return NULL;
 }
 
+void UpdatePunc(FcitxSunpinyin* sunpinyin)
+{
+    if (!sunpinyin->puncOp)
+        return;
+    const char symbol[] = "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~";
+    string_pairs puncPairs;
+    for( int i = 0; i < sizeof(symbol)/sizeof(char) -1; i++) {
+        int c = symbol[i];
+        char s[2] = {symbol[i], '\0'};
+        char* p1 = NULL, *p2 = NULL;
+        FcitxModuleFunctionArg args;
+        args.args[0] = &c;
+        args.args[1] = &p1;
+        args.args[2] = &p2;
+        InvokeFunction(sunpinyin->owner, FCITX_PUNC, GETPUNC2, args);
+        string_pair p;
+        p.first = s;
+        if (p1) {
+            p.second = p1;
+            puncPairs.push_back(p);
+        }
+        if (p2) {
+            p.second = p2;
+            puncPairs.push_back(p);
+        }
+    }
+    sunpinyin->puncOp->initPunctMap(puncPairs);
+}
+
+void* SunpinyinAddWord(void* arg, FcitxModuleFunctionArg args)
+{
+    FcitxSunpinyin* sunpinyin = (FcitxSunpinyin*) arg;
+
+    char* word = (char*) args.args[0];
+
+    CUserDict* dict = sunpinyin->view->getIC()->getUserDict();
+
+    if (!dict)
+        return NULL;
+
+    IPySegmentor::TSegmentVec& segments = sunpinyin->view->getPySegmentor()->getSegments(false);
+    if (segments.size() == 0)
+        return NULL;
+    CSyllables syls;
+    size_t len = fcitx_utf8_strlen(word);
+
+    /* sunpinyin have assert inside add word, this is must be checked first */
+    if (len > MAX_USRDEF_WORD_LEN || len < 2)
+        return NULL;
+
+    /* no way to check real single character pronouce, but let it be here */
+    for (int i = 0; i < segments.size(); i ++) {
+        const IPySegmentor::TSegment& segment = segments[i];
+        for (int j = 0; j < segment.m_syllables.size(); j ++) {
+            TSyllable syl = segment.m_syllables[j];
+            if (!syl.isFullSyllable())
+                return NULL;
+            syls.push_back(syl);
+        }
+    }
+
+    /* check the size */
+    if (syls.size() != fcitx_utf8_strlen(word))
+        return NULL;
+
+    TWCHAR* wword = (TWCHAR*) fcitx_utils_malloc0(sizeof(TWCHAR) * (len + 1));
+    MBSTOWCS(wword, word, len);
+    dict->addWord(syls, wword);
+    free(wword);
+
+    return NULL;
+}
+
 // kate: indent-mode cstyle; space-indent on; indent-width 0;
